
_ATA_INC_	.SET	1
_DMA_INC_	.SET	1

	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc
	.INCLUDE inc\DPDSK.INC

.LIST on

	;.INCLUDE inc\BIOSSEG.INC

	
WRITETRASF	.EQU	$04	; write transfer: read ATA, write MEM
READTRASF	.EQU	$08	; read  transfer: read MEM, write ATA

BUSYTMO		.EQU	2000	; 2000ms (timer PB6) - busy timeout
INTTMO		.EQU	2000	; 2000ms (timer PB6) - interrupt timeout
T2MS5		.EQU	((2500 * PHI2) - 1)	; timeout 2.5ms
T2MS10		.EQU	((10000 * PHI2) - 1)	; timeout 10ms

XBMIN		.EQU	$0B40	; min. x-mem buffer
XBMAX		.EQU	$1000	; max. x-mem buffer + 1
MBMIN		.EQU	$01	; min. buffer bank
MBMAX		.EQU	$F0	; max. buffer bank + 1
MDMAX		.EQU	$0800	; max. dma buffer

READSECTOR	.EQU	$20	; read sector(s)
WRITESECTOR	.EQU	$30	; write sector(s)
READDMA		.EQU	$C8	; read dma sectors
WRITEDMA	.EQU	$CA	; write dma sectors

;FLUSHCACHE	.EQU	$E7
IDENTDEV	.EQU	$EC	; identify device

; page 0 local var's (declared in bios temp. work area)
_P0BTMP:	.SECTION page0, ref_only, offset bwrktmp	; DP Tmp. BIOS
ATATMP_START	.EQU	$

drive		.DB		; ata #0 or ata #1 device
lba		DS	3	; 24 bit sector address
secnum		.DB		; sector(s) count # (if 0 -> 256 sec.)
seccnt		.DB		; copy of secnum
ifrbit		.DB		; bit to test on VIA3 IFR
cmdfg		.DB		; flag for load command block reg's
ataop		.DB		; ata operation flag
				; <7>=0->read, 1->write
				; <6>=1 compare (ignored in dma mode)
memfg		.DB		; using buffer in x-mem flag
drdyf		.DB		; test alt. status reg. while read/write block
match		.DB		; compare result (match if not null)
bufp		LP		; in/out buffer long pointer
memp		LP		; start buffer pointer
xbuf		.DW		; x-mem buffer # ($0B40..$0FFF)
				; dma buffer # ($0000..$7FFF)
xbnk		.DB		; starting x-mem bank
xsave		.DB		; save x-mem setting
nlba		.DW		; total sector's count #
ncnt		.DW		; dma words count #
tmp		.DB		; temp
ataofs		.DB		; temp: ata offset address
wtmp		.DW		; temp. word
mbrsgn		.DW		; mbr signature
mbrfg		.DW		; mbr first partition flag
tmpy		.DB
dmareq		.DB

ATATMP_SIZ	.EQU	($-ATATMP_START)
	.ENDS

;---------------------------------------------------------------------------
; code segment -- bank $F8
;---------------------------------------------------------------------------

	.CODEF8
	
	.LONGA	off
	.LONGI	off
	
;---------------------------------------------------------------------------
; System interface
;---------------------------------------------------------------------------

;---------------------------------------------------------------------------
; sys_atadma - read/write sectors in ata device (dma mode)
;
; prototype: sys_atadma(bDev, dwSec, bNum, nBuf)
;
; Params (7 bytes):
;	bDev	= ata device # (0..1)
;	dwSec	= 24 bit address of first sector (LBA)
;	bNum	= number of sectors to read/write; if bNum=0 then bNum=256
;	nBuf	= dma buffer # ($0000..$07FF)
;
; Out:
;	CF = 0 if operation completed successfully
;		C  = uncahnged
;		X  = unchanged
;		Y  = unchanged
;	CF = 1 if error
;		C  = unchanged
;		X  = extended error code (if any)
;		Y  = error code
;	Status register always preserved except carry

; params offset
nBuf	.SET	STKPRMS
bNum	.SET	STKPRMS + 2
dwSec	.SET	STKPRMS + 3
bDev	.SET	STKPRMS + 6

sys_atadma:
	.GLOBAL	sys_atadma

	stz	ataerr		; clear error var's
	stz	ataxer
	ACC16			; retrieve function number
	lda	STKPCL,s	; pointer to byte after signature
	sta	COPPtr
	inc	a		; update return address
	sta	STKPCL,s
	lda	dwSec,s		; get 16 bit param's
	sta	lba
	lda	nBuf,s
	cmp	#MDMAX
	bcs	?11		; buffer overflow 
	sta	xbuf
	ACC08			; A,M -> 8 bit
	lda	STKPBR,s	; bank where was executed cop instruction
	sta	COPPtr+2	
	lda	[COPPtr]	; byte after signature
	cmp	#$02		; right value: 0,1
	bcc	?00		; ok
	lda	#ATA_BADOP	; invalid function
	bra	?20		; set error & exit
?00:	tax			; X=operation (0=read,1=verify,2=write)
	lda	>OPTAB2,x
	sta	ataop		; operation ($00, $80)
	lda	dwSec+2,s
	sta	lba+2		; store first sector lba
	ldy	#0
	lda	bNum,s
	sta	secnum		; sector's count #
	sta	wtmp		; for check
	bne	?02		; less than 256
	iny			; = 256
?02:	sty	wtmp+1		; for later check
	lda	bDev,s		; device
	and	#1		; 0 or 1
	tax
	bit	atadev,x	; ready?
	bmi	?03		; yes
	lda	#ATA_NOTREADY	; not ready
	bra	?20		; set error & exit
?03:	stx	drive
	lda	wtmp
	sta	ncnt+1		; ncnt = words count
	stz	ncnt
	ACC16CLC
	dec	ncnt		; dma need words count - 1
	lda	xbuf		; buffer #
	adc	wtmp		; add sec. #	
	cmp	#MDMAX+1	; buffer overflow?
	bcs	?11		; yes
	dec	a		; final buffer
	ACC08
	xba			; transfer must be in one 64Kw bank
	cmp	xbuf+1
	bne	?12		; error: not same bank
	jsr	dmardwr		; execute dma transfer
	bcs	?20		; error
	rtl
?11:	ACC08
?12:	lda	#ATA_MEM	; bad buffer
?20:	tax
	lda	#0		; report extended error if any
	sta	STKXR+1,s
	cpx	#ATA_EBIT
	bne	?22
	tya			; extended error
?22:	sta	STKXR,s
	sta	ataxer		; extended error code
	txa			; A=error code
	sta	ataerr
	sec
	rtl

;---------------------------------------------------------------------------
; sys_atarw - read/write/verify sectors in ata device (pio mode)
;
; prototype: sys_atarw(bDev, dwSec, bNum, lpBuf, bFlag)
;
; Params (9 bytes):
;	bDev	= ata device # (0..1)
;	dwSec	= 24 bit address of first sector (LBA)
;	bNum	= number of sectors to read/write; if bNum=0 then bNum=256
;	lpBuf	= long pointer to buffer
;			if bank(lpBuf+2)=0 or bank=1 buffer is in the x-mem
;			and in this case lpBuf store the x-mem buffer;
;			x-mem buffer can be only in range $0B40..$0FFF
;			if conventional memory, bank must be in range $01..$EF
;	bFlag	= flag
;			<7>: verify data after writing
;			ignored in read/verify operation
;
; Out:
;	CF = 0 if operation completed successfully
;		C  = uncahnged
;		X  = unchanged
;		Y  = unchanged
;	CF = 1 if error
;		C  = unchanged
;		X  = extended error code (if any)
;		Y  = error code
;	Status register always preserved except carry

; params offset
bFlag	.SET	STKPRMS
lpBuf	.SET	STKPRMS + 1
bNum	.SET	STKPRMS + 4
dwSec	.SET	STKPRMS + 5
bDev	.SET	STKPRMS + 8

sys_atarw:
	.GLOBAL	sys_atarw

	stz	ataerr		; clear error var's
	stz	ataxer
	ACC16			; retrieve function number
	lda	STKPCL,s	; pointer to byte after signature
	sta	COPPtr
	inc	a		; update return address
	sta	STKPCL,s
	lda	dwSec,s		; get 16 bit param's
	sta	lba
	lda	lpBuf,s
	sta	memp
	ACC08			; A,M -> 8 bit
	lda	STKPBR,s	; bank where was executed cop instruction
	sta	COPPtr+2	
	lda	[COPPtr]	; byte after signature
	cmp	#$03		; right value: 0,1,2
	bcc	?00		; ok
	lda	#ATA_BADOP	; invalid function
	jmp	?20		; set error & exit
?00:	tax			; X=operation (0=read,1=verify,2=write)
	lda	>OPTAB,x
	bpl	?01		; read/verify (ignore bFlag)
	tay
	lda	bFlag,s		; verify flag (after write)?
	bmi	?00a		; yes
	tya
	bra	?01		; no
?00a:	tya	
	ora	#$40		; yes, set verify flag
?01:	sta	ataop		; operation ($00, $40, $80, $C0)
	lda	dwSec+2,s
	sta	lba+2		; store first sector lba
	ldy	#0
	lda	bNum,s
	sta	secnum		; sector's count #
	sta	wtmp		; for check
	bne	?02		; less than 256
	iny			; = 256
?02:	sty	wtmp+1		; for later check
	lda	bDev,s		; device
	and	#1		; 0 or 1
	tax
	bit	atadev,x	; ready?
	bmi	?03		; yes
	lda	#ATA_NOTREADY	; not ready
	bra	?20		; set error & exit
?03:	stx	drive
	stz	memfg
	lda	lpBuf+2,s	; buffer bank
	cmp	#MBMIN		; x-mem buffer?
	bcs	?08		; no, conventional memory
	ACC16
	ldy	#0		; assume no error
	lda	memp		; check x-mem buffer
	sta	xbuf
	cmp	#XBMIN
	bcc	?04		; error
	clc
	adc	wtmp		; add sector's count
	bcs	?04		; overflow error
	cmp	#XBMAX+1
	bcc	?06		; buffer limits ok
?04:	ldy	#ATA_XMEM	; Y=x-mem buffer error
?06:	ACC08
	tya
	bne	?20		; set error & exit
	lda	#$80
	sta	memfg
	bra	?14		; ok, good x-mem buffer
?08:	sta	memp+2		; store buffer bank
	cmp	#MBMAX		; over max. bank?
	bcs	?12		; yes, error
	tax			; X=bank
	ACC16
	stz	nlba+2		; high word (only low byte significative)
	lda	wtmp		; convert secnum in bytes
	ldy	#9		; 9 shift (mult by 512)
?10:	asl	a		; low word
	rol	nlba+2		; shift high word
	dey
	bne	?10
	clc
	adc	memp
	sta	nlba		; end address: low word
	ACC08
	txa			; start bank
	adc	nlba+2		; end address: bank
	cmp	#MBMAX
	bcc	?14		; ok, good buffer
?12:	lda	#ATA_MEM	; bad buffer
	bra	?20		; set error & exit
?14:	bit	ataop
	bmi	?18		; write op
	jsr	readsct
	bcs	?20		; read error
?16:	rtl			; ok, no error
?18:	jsr	writesct
	bcc	?16		; write ok
?20:	tax
	lda	#0		; report extended error if any
	sta	STKXR+1,s
	cpx	#ATA_EBIT
	bne	?22
	tya			; extended error
?22:	sta	STKXR,s
	sta	ataxer		; extended error code
	txa			; A=error code
	sta	ataerr
	sec
	rtl

;---------------------------------------------------------------------------
; sys_atactl - control function 
;
; prototype: sys_atactl(A=function#, X=device#)
;
; Functions:
;	A=$00: 	reset device X
;	A=$01: 	identify device X
;	A=$02: 	load mbr of device X
;	A=$03: 	load pbr of device X
;
; Out:
;	CF = 0 if operation completed successfully
;	CF = 1 if error
;		C  = unchanged
;		X  = extended error code (if any)
;		Y  = error code
;	Status register always preserved except carry
sys_atactl:
	.GLOBAL	sys_atactl

	stz	ataerr		; clear error var's
	stz	ataxer	
	lda	STKXR,s
	and	#1
	sta	drive
	tax
	bit	atadev,x
	bpl	?10		; not ready
	txa
	beq	?02
	ldx	#ATAOFS
?02:	stx	ataofs
	lda	STKCR,s		; function#
	cmp	#$04
	bcs	?12		; invalid function
	tay
	bne	?04	
	jsr	devres		; 0: reset device
	bra	?08
?04:	ldx	drive	
	dec	a
	beq	?06		; 1: ident
	dec	a
	beq	?05		; 2: load mbr
	jsr	getpbr		; 3: load pbr
?05:	lda	#11100010B
	and	atadev,x
	sta	atadev,x	
	jsr	chkmbr		; 2: load mbr
	bra	?08
?06:	lda	#11011111B	; 1: ident
	and	atadev,x
	sta	atadev,x
	jsr	ident
?08:	bcs	?14
	rtl
?10:	lda	#ATA_NOTREADY	; not ready
	bra	?14		; set error & exit
?12:	lda	#ATA_BADOP	; invalid function
?14:	tax
	lda	#0		; report extended error if any
	sta	STKXR+1,s
	cpx	#ATA_EBIT
	bne	?16
	tya			; extended error
?16:	sta	STKXR,s
	sta	ataxer		; extended error code
	txa			; A=error code
	sta	ataerr
	sec
	rtl

;---------------------------------------------------------------------------
; basic read/write routines
;---------------------------------------------------------------------------

	.LONGA	off
	.LONGI	off

; write N (0..256) sector(s) on ata device 
; in:	drive=0 or 1 (ata #0 or ata #1 device)
;	lba[0..2]=24 bits starting sector address
;	secnum=sector(s) count (if 0 -> 256 sect.)
;	memp=long pointer to input buffer
;	memfg<7>=flag buffer x-mem
;	xbuf=x-mem buffer # if using x-mem ($0B40..$0FFF)
;	ataop<6>=flag compare data after writing
; Out:	CF=0 if no error
;	CF=1 if error, A=error code, Y=error reg.
; use:	all
writesct:
	jsr	setbuf		; set buffer pointer
	lda	#$80
	sta	cmdfg		; load command block registers
	lda	#$80
	trb	ataop		; we know is a write op...
	ldy	#WRITESECTOR	; write sectors PIO mode
	jsr	sndcmd		; return X=ata offset
	nop			; delay before to access status reg. (400ns)
	nop
	nop
	nop
	;bra	?w1	
	bit	!ATAALTST,x	; dummy read	
?w0:	lda	!ATAST,x	; read status reg.
	bmi	?w0		; wait BUSY bit=0
	bit	#ATADFB+ATAERRB
	bne	?idle		; error
	bit	#ATADRQ		; ready to transfer?
	bne	?wrb		; yes, write data block
	bra	?w0		; wait for DRQ
?loop:	bit	!ATAALTST,x	; dummy read
?w1:	lda	!ATAST,x	; read status reg.
	bmi	?w1		; wait BUSY bit=0
	bit	#ATADRQ		; end of transfer or error?
	beq	?idle		; yes
	ldy	seccnt
	bne	?wrb		; more blocks
	lda	#$40
	tsb	drdyf		; data loss
	inc	seccnt
?wrb:	jsr	wrblk		; write data block (256 words)
	dec	seccnt		; update left count #
?int:	jsr	waitint		; wait interrupt assertion within timeout
	bcc	?loop		; ok, read status reg.

	; no int. assertion: this happen if device expect more data block
	; write out data block until DRQ bit reset
	lda	#$40
	tsb	drdyf		; signal this error
?w2:	lda	!ATAALTST,x
	bit	#ATADRQ
	beq	?w3		; no drq	
	stz	!ATAPORT,x	; latch data high byte
	stz	!ATAPORT+1,x	; latch data high byte
	stz	!ATADATA,x	; write
	bra	?w2
?w3:	lda	!ATAALTST,x
	lda	!ATAALTST,x
	lda	!ATAST,x
	bmi	?int

?idle:	bit	drdyf		; test interrupt timeout error
	bvc	?chk		; no, so test error condition
	jsr	devres		; reset device because write error
	lda	#ATA_DLOSS	; data loss while writing
	bra	?err
?chk:	bit	#ATADFB		; device fault bit ?
	beq	?chk1		; no
	lda	#ATA_FAULT	; yes
	bra	?err
?chk1:	bit	#ATAERRB	; error bit ?
	beq	?chk2		; no...so write ok
	ldy	!ATAERROR,x	; read error register
	lda	#ATA_EBIT	; error bits
	bra	?err
?chk2:	lda	seccnt		; all blocks was written?
	beq	?ok		; yes
	jsr	devres		; reset device because premature ending
	lda	#ATA_DEVRES
	bra	?err
?ok:	clc			; assume no error	
	bit	ataop		; request for verify data?
	bvc	?done		; no, exit (no error)
	jsr	cmpsct		; compare written data
	bcc	?done		; ok, no error
	;lda	#ATA_VERF	; assume always data verify error
?err:	sec			; error flag
?done:	bit	memfg		; using x-mem?
	bpl	?rts		; no
	ldx	xsave
	sta	!CRXME,x	; restore
?rts:	rts

; load MBR in memory
; in:	drive
getmbr:
	ACC16			; set pointer for MBR
	lda	#MBR0ATA
	ldx	drive
	beq	?02
	lda	#MBR1ATA
?02:	sta	memp
	sta	bufp
	stz	lba		; read sector=0
	ACC08
	lda	#^MBR0ATA
	sta	memp+2
	sta	bufp+2
	stz	lba+2
	stz	ataop		; just read
	stz	memfg		; no x-mem
	lda	#1
	sta	secnum		; read just one block
	
	jmp	_dmald
	
	bra	readsct2

; load PBR in memory
; in:	drive
getpbr:
	lda	drive
	asl	a
	asl	a
	tax			; index to access atambr var
	ACC16			; set pointer for MBR
	lda	#PBR0ATA
	ldy	drive
	beq	?02
	lda	#PBR1ATA
?02:	sta	memp
	sta	bufp
	lda	atambr,x	
	sta	lba		; read pbr sector
	ACC08
	lda	#^MBR0ATA
	sta	memp+2
	sta	bufp+2
	lda	atambr+2,x
	sta	lba+2
	stz	ataop		; just read
	stz	memfg		; no x-mem
	lda	#1
	sta	secnum		; read just one block
	
	jmp	_dmald

	bra	readsct2

; get identification device data
; in:	drive
getid:
	ACC16			; set pointer for identification data
	lda	#ID0ATA
	ldx	drive
	beq	?02
	lda	#ID1ATA
?02:	sta	memp
	sta	bufp
	ACC08
	lda	#^ID0ATA
	sta	memp+2
	sta	bufp+2
	stz	cmdfg		; no load command block registers
	stz	ataop		; just read
	stz	memfg		; no x-mem
	lda	#1
	sta	secnum		; read just one block
	ldy	#IDENTDEV	; identify device command
	bra	piord		; pio mode read
	
; compare buffer with ata device after writing
cmpsct:
	lda	memp		; restore buffer
	sta	bufp
	lda	memp+1
	sta	bufp+1
	lda	memp+2
	sta	bufp+2
	bit	memfg		; using x-mem?
	bpl	readsct2	; no
	lda	#$40
	trb	memfg		; no restore x-mem setting when finish
	lda	xbnk
	sta	!PIA0+PIAPRA	; set bank
	bra	readsct2

; read/compare N (0..256) sector(s) on ata device 
; in:	drive=0 or 1 (ata #0 or ata #1 device)
;	lba[0..2]=24 bits starting sector address
;	secnum=sector(s) count (if 0 -> 256 sect.)
;	memp=long pointer to input buffer
;	memfg<7>=flag buffer x-mem
;	memfg<6>=restore x-mem setting on exit
;	xbuf=x-mem buffer # if using x-mem ($0B40..$0FFF)
;	ataop<6>=flag compare data
; Out:	CF=0 if no error
;	CF=1 if error, A=error code, Y=error reg.
; use:	all
readsct:
	bit	memfg		; using x-mem?
	bpl	?02		; no
	lda	#$40
	tsb	memfg		; restore x-mem setting when finish
?02:	jsr	setbuf		; set buffer pointer

; entry point for compare after writing
readsct2:	
	lda	#$80
	sta	cmdfg		; load command block registers
	sta	match		; good compare
	ldy	#READSECTOR	; read sector(s) command

; entry point for identify device
piord:
	jsr	sndcmd		; return X=ata offset
	lda	ifrbit		; check interrupt
?w1:	bit	VIA3+VIAIFR	; test CA1/CB1 transition
	beq	?w1		; wait interrupt assertion from ata device
	sta	VIA3+VIAIFR	; clear interrupt flag
?w2:	lda	!ATAST,x	; wait BUSY bit=0
	bmi	?w2
	bit	#ATADRQ		; ready to read a data block?
	bne	?rdb		; yes, read first data block
	bit	#ATADFB+ATAERRB
	bne	?chk		; error -- stop
	bra	?w2		; some CF need time to set DRQ 	
?loop:	jsr	waitint		; wait interrupt assertion within timeout
	bcc	?w4		; ok, read status reg.
	lda	#$40		; no interrupt assertion
	tsb	drdyf		; signal this error
?w3:	lda	!ATAALTST,x
	bit	#ATADRQ		; device expect more data block to transfer?
	beq	?w30		; no...continue... keep error condition 	
	lda	!ATADATA,x	; read wrong data and discard...
	bra	?w3		; ...until end of data block
?w30:	lda	!ATAALTST,x
	lda	!ATAST,x
	bpl	?tst
	brk
?w4:	lda	!ATAST,x
	bmi	?w4
	bit	#ATADRQ		; device expect more data block to transfer?
	beq	?tst		; no... check error condition		
?rdb:	jsr	rdblk		; read a block (256 words)	
?nxt:	dec	seccnt		; update left blocks count #
	bne	?loop		; continue until all sectors was readden
?idle:	lda	!ATAST,x	; read status reg.
	bpl	?tst
	brk
?tst:	bit	drdyf		; test interrupt timeout error
	bvc	?chk		; no, so test error condition
	jsr	devres		; reset device because read error
	lda	#ATA_DEVRES	; data loss while reading
	lda	#$FF
	bra	?err	
?chk:	bit	#ATADFB		; device fault bit ?
	beq	?chk1		; no
	lda	#ATA_FAULT	; yes
	bra	?err		; error
?chk1:	bit	#ATAERRB	; error bit ?
	beq	?chk2		; no... all seem good
	ldy	!ATAERROR,x	; read error register
	lda	#ATA_EBIT	; error bits
	bra	?err
?chk2:	ldy	seccnt
	beq	?chk3
	jsr	devres		; reset device because read error
	lda	#ATA_DLOSS	; data loss while reading
	bra	?err
?chk3:	clc			; assume no error	
	bit	ataop		; was compare data operation?
	bvc	?done		; no, exit (no error)
	ldy	match		; all match?
	bne	?done		; ok, good compare
	lda	#ATA_VERF	; compare error
?err:	sec			; error flag
?done:	bit	memfg		; using x-mem?
	bpl	?rts		; no
	bvc	?rts		; no restore x-mem setting
	ldx	xsave
	sta	!CRXME,x	; restore
?rts:	rts

; load DMA mbr/pbr
_dmald:
	stz	xbuf
	stz	xbuf+1
	stz	ncnt+1
	lda	#$FF
	sta	ncnt
	jsr	dmardwr
	bcs	?100
	lda	#$08		; PB3 VIA3
	trb	VIA3+VIAPRB	; enable 1Mb buffer
	lda	!VIA3+VIAPRANH
	and	#$F8
	sta	!VIA3+VIAPRANH	; set bank

	ldy	#0
	stz	bufp
	stz	bufp+1
	lda	#2
	sta	bufp+2
	INDEX16
	ldx	#$200
?10:	lda	[bufp],y
	sta	[memp],y
	iny
	dex
	bne	?10
	INDEX08
	clc
?100:	rts

; read/write N (0..256) sector(s) on ata device in dma mode 
; in:	drive = 0 or 1 (ata #0 or ata #1 device)
;	ataop = ata operation flag: <7>=0->read, 1->write
;	lba[0..2] = 24 bits starting sector address
;	secnum = sector(s) count (if 0 -> 256 sect.)
;	xbuf = dma buffer #
;	ncnt = words count # - 1
; Out:	CF=0 if no error
;	CF=1 if error, A=error code, Y=error reg.
; use:	all
dmardwr:
	lda	#$80
	sta	cmdfg		; load command block registers
	ldx	drive
	lda	>DMAREQ,x
	sta	wtmp+1		; check dmareq line
	jsr	dmaset		; dma setup	
	ldy	#READDMA	; read sector(s) command
	bit	ataop
	bpl	?do
	ldy	#WRITEDMA	; write sector(s) command
?do:	jsr	sndcmd		; return X=ata offset
	stz	seccnt
	nop
	nop
	nop
	nop
	lda	>ATAALTST,x
?w0:	lda	>ATAST,x	; wait
	and	#ATABSY+ATADRQ
	beq	?chk		; end - check error
	
	;lda	!VIA3+VIAPRANH
	;and	wtmp+1		; dma request?
	;beq	?w0		; no
	lda	>DMA0SR
	and	dmareq
	beq	?w0

	lda	#$02		; enable dma
	trb	VIA0+VIAPRANH

	;nop
	;nop
	;nop
	;nop
	;lda	ifrbit		; check interrupt
;?w1:	bit	VIA3+VIAIFR	; test CA1/CB1 transition
;	beq	?w1		; wait interrupt assertion from ata device
;	sta	!VIA3+VIAIFR	; clear interrupt flag

	jsr	waitint
	lda	#$02		; disable dma
	tsb	VIA0+VIAPRANH
	bcc	?tst		; seem good: error test
	
	; no interrupt whitin timeout
	lda	>DMA0SR		; clear status flag
	sta	>DMA0WMCLR	; reset dma controller
	bra	?res		; data loss
	
?tst:	lda	>ATAST,x	; clear interrupt	
?chk:	bit	#ATADFB		; device fault bit ?
	beq	?chk1		; no
	lda	#ATA_FAULT	; yes
	bra	?err		; error
?chk1:	bit	#ATAERRB	; error bit ?
	beq	?chk2		; no... all seem good
	ldy	!ATAERROR,x	; read error register
	lda	#ATA_EBIT	; error bits
	bra	?err
?chk2:
	; check TC on dma status
	lda	wtmp
	ldx	drive
	and	>DMACHNL,x
	beq	?loss		; data loss: no TC
	
	;ldy	seccnt
	;beq	?chk3
	bne	?chk3

?res:	jsr	devres		; reset device because read error
?loss:	lda	#ATA_DLOSS	; data loss while reading
	bra	?err
?chk3:	clc			; assume no error	
	rts	
?err:	sec			; error flag
	rts

waitint:
	lda	#$20		; set T2 count PB6 pulse (1ms)
	tsb	VIA0+VIAACR	
	lda	#<INTTMO	; interrupt timeout
	sta	!VIA0+VIAT2CL
	lda	#>INTTMO
	sta	!VIA0+VIAT2CH
	sei
?10:	lda	ifrbit		; flag bit for test CA1/CB1 
	bit	VIA3+VIAIFR
	bne	?20		; interrupt assertion (exit with CF=0)
	lda	#T2IFRB
	bit	VIA0+VIAIFR
	beq	?10		; wait for INTTMO ms
	lda	ifrbit		; flag bit for test CA1/CB1 
	bit	VIA3+VIAIFR
	bne	?20		; interrupt assertion
	cli
	sec			; timeout
	rts
?20:	sta	VIA3+VIAIFR	; clear interrupt flag
	cli
	clc
	rts

;---------------------------------------------------------------------------
; low level routines
;---------------------------------------------------------------------------

	.LONGA	off
	.LONGI	off

; read/compare data block (512 bytes) from ata device
; in:	X=offset ata address
;	ataop<6>=1 if compare, else read
;	match=not null
;	bufp=long pointer to input buffer
;	drdyf<7>=1 if need test DRDY bit in alternate status
;	memfg<7>=flag buffer in x-mem
;out:	bufp updated
;	match=0 if not match
; use:	A,Y
rdblk:
	php			; save I flag
	sei			; disable interrupt
	ldy	#0		; buffer index
	INDEX16			; X,Y 16 bit - XH,YH cleared
	bit	ataop		; test bit 6: compare?
	bvs	?12		; yes
	
	ACC16
?02:	lda	>ATADATA,x	; latch data
	lda	!ATAPORT,x	; read latched data
	sta	[bufp],y
	iny
	iny
	;bit	drdyf		; test DRDY bit in alternate status?
	;bpl	?06		; no
?04:	;bit	!ATAALTST-1,x	; wait DRDY bit high
	;bvc	?04
?06:	cpy	#$0200			
	bcc	?02		; loop 512 bytes	
	ACC08
	bra	updp		; update bufp pointer

	; compare data block
?12:	ACC16
	lda	>ATADATA,x	; read data low, latch data high
	lda	!ATAPORT,x	; read latched data
	cmp	[bufp],y
	ACC08	
	beq	?14
	stz	match		; no match
?14:	iny
	iny
	;bit	drdyf		; test DRDY bit in alternate status?
	;bpl	?18		; no
?16:	;bit	!ATAALTST,x	; wait DRDY bit high
	;bvc	?16
?18:	cpy	#$0200			
	bcc	?12		; loop 512 bytes	
	bra	updp		; update bufp pointer

; write data block (512 bytes) on ata device
; In:	X=offset ata address
;	bufp=long pointer to output buffer
;	drdyf<7>=flag test DRDY bit in alternate status
;	memfg<7>=flag buffer in x-mem
; out:	bufp updated
; use:	A,Y
wrblk:
	.LONGA	off
	.LONGI	off

	php			; save I flag
	sei			; disable interrupt
	ldy	#0		; buffer index
	
	INDEX16			; X,Y 16 bit - XH,YH cleared

	;lda	#$20			; /MW0 high
	;trb	VIA1+VIAPRB

	ACC16
?02:	;ACC16
	lda	[bufp],y	; read data
	sta	>ATAPORT,x	; latch data
	;ACC08
	
	lda	#0			; /MW0 low
	tsb	!0
	lda	#0			; /MW0 low
	tsb	!0
	lda	#0			; /MW0 low
	tsb	!0

	sta	>ATADATA,x	; send data low + latched data high	
	iny
	iny
	;lda	#$20			; /MW0 high
	;bit	$ff
	;bpl	?04
	;tsb	VIA1+VIAPRB
	
	;bit	drdyf		; test DRDY bit in alternate status?
	;bpl	?06		; no
?04:	;lda	>ATAALTST-1,x
	;bit	!ATAALTST,x
	lda	>ATAALTST,x
	;bmi	?04
	;bvc	?04
	;asl	a
	;bpl	?04		; wait DRDY bit high

?06:	cpy	#$0200			
	bcc	?02		; loop 512 bytes
	ACC08
	;lda	#$20			; /MW0 high
	;tsb	VIA1+VIAPRB
	
; update bufp pointer after read/write data block
updp:
	.LONGI	on

	ldy	bufp+1		; update bufp
	iny
	iny
	sty	bufp+1
	bit	memfg		; buffer in x-mem?
	bpl	?10		; no, all done
	ldy	bufp
	cpy	#XMSTOP		; still in x-mem window?
	bcc	?10		; yes, all done
	ldy	#XMSTART	; restart at x-mem beginning window...
	sty	bufp
	inc	!PIA0+PIAPRA	; ...and at next x-mem bank 
?10:	CPU08
	plp			; restore I flag
	rts

; enable 8bit transfer
set8:
	lda	#$01
	bra	setmode

set16:
	lda	#$81

setmode:
	pha
	ldx	drive		; ata device #
	lda	>IFRTST,x
	sta	ifrbit		; save right IFR bit to test
	txa
	beq	?02		; offset of ata 0
	ldx	#ATAOFS		; offset of ata 1
?02:	stz	!ATACTRL,x	; enable ata device to assert interrupt
	pla
	sta	!ATAFEAT,x
	lda	ifrbit	
	sta	!VIA3+VIAIFR	; clear CA1/CB1 flag in IFR
	lda	#$EF		; command
	sta	!ATACMD,x	; store command
	lda	ifrbit
	sei
?10:	bit	!VIA3+VIAIFR
	beq	?10
	cli
	lda	!ATAST,x
	lda	!ATAERROR,x
	sta	$ff
	rts

; send a command to ata device
; after sending a command will wait for interrupt
; in:	Y=command	
;	drive=0 or 1 (ata #0 or ata #1 device)
;	cmdfg<7>=flag load command block registers
;	lba[0..2]=24 bits starting lba (if need)
;	secnum=sector(s) count (if need)
; out:	X=offset ata address
;	irfbit=bit for test CA1/CB1 on VIA 1 (int. line for ata0/1)
;	seccnt=sector(s) count while read/write block
; use:	A,X,Y
sndcmd:
	lda	secnum
	sta	seccnt		; save sector(s) count #
	ldx	drive		; ata device #
	lda	>IFRTST,x
	sta	ifrbit		; save right IFR bit to test
	stz	drdyf
	bit	atadev,x	; set drdy flag for C.F.
	bvc	?01
	lda	#$80
	sta	drdyf
?01:	txa
	beq	?01a		; offset of ata 0
	ldx	#ATAOFS		; offset of ata 1
?01a:
	lda	#$40
	sta	>ATADEV,x

	;stz	!ATADEV,x
	nop
?01c:	bit	!ATAALTST,x	: wait device ready
	bmi	?01c
	bvc	?01c
	
	lda	#0
	sta	>ATACTRL,x	; enable ata device to assert interrupt
	
	;stz	!ATACTRL,x	; enable ata device to assert interrupt
	nop
	bit	cmdfg
	bpl	?15
	
	;lda	#0		; dev. reg. for command that not need params
	;sta	!ATADEV,x
	;bit	cmdfg
	;bpl	?02a
?02:	lda	#$40		; device 0 & lba is limited to 24 bits
	sta	>ATADEV,x	; store dev. register
	;nop
	;nop
	;nop
	;nop
?02a:	bit	!ATAALTST,x	: wait device ready
	bmi	?02a
	bvc	?02a
	;stz	!ATACTRL,x	; enable ata device to assert interrupt
	;nop
	;nop
	;lda	#0		; dev. reg. for command that not need params
	;bit	cmdfg
	;bpl	?04
	;lda	#$E0		; device 0 & lba is limited to 24 bits
	;lda	#$40
?04:	;sta	>ATADEV,x	; store dev. register
?06:	;sta	>ATADEV,x
	;bit	!ATAALTST,x	; wait device ready
	;bmi	?06
	;bvc	?06
	;bit	cmdfg		; will load command block register's?
	;bpl	?16		; no
	lda	seccnt		; store sectors count
	sta	>ATASCTCNT,x
?08:	;bit	!ATAALTST,x	; wait device ready
	;bmi	?08
	;bvc	?08
	lda	lba		; store starting sector
	sta	>ATALBAL,x
?10:	;bit	!ATAALTST,x	; wait device ready
	;bmi	?10
	;bvc	?10
	lda	lba+1
	sta	>ATALBAM,x
?12:	;bit	!ATAALTST,x	; wait device ready
	;bmi	?12
	;bvc	?12
	lda	lba+2
	sta	>ATALBAH,x
?14:	;bit	!ATAALTST,x	; wait device ready
	;bmi	?14
	;bvc	?14
	lda	#$40
	sta	>ATADEV,x	; store dev. register
?15:	bit	!ATAALTST,x	; wait device ready
	bmi	?15
	bvc	?15

?16:	lda	ifrbit	
	sta	!VIA3+VIAIFR	; clear CA1/CB1 flag in IFR
	tya			; command
	sta	!ATACMD,x	; store command
	rts

; set buffer pointer
; use: A,B,X
setbuf:
	ldx	memp+2
	bit	memfg		; using x-mem?
	ACC16
	bpl	?10		; no, just set bufp=memp
	lda	xbuf		; x-mem buffer #
	asl	a		; x16 -> B=x-mem bank
	asl	a
	asl	a
	asl	a
	xba			; 00xx,10xx,20xx,...,E0xx,F0xx
	tax			; X=x-mem bank
	stx	xbnk		; save starting bank
	stx	PIA0+PIAPRA	; set bank
	and	#$F000		; mask off
	lsr	a		; 00xx,02xx,04xx,...,1Cxx,1Exx
	lsr	a
	lsr	a		; 0000,0200,0400,...,1C00,1E00
	ora	#XMSTART	; point to starting buffer
	sta	memp
	ldx	CRXME		; save x-mem setting
	stx	xsave
	stx	CRXMEON		; enable x-mem in window $012000 - $013FFF
	ldx	#XMBANK
	stx	memp+2
	bra	?20
?10:	lda	memp		; set bufp=memp
?20:	sta	bufp
	stx	bufp+2
	ACC08
	rts

; software reset on ata device
; in:	X=offset ata address
devres:
	lda	#CTRLSRST	; software reset control
	sta	!ATACTRL,x
	jsr	?40		; delay 16uS (need at least 5uS)
	stz	ATACTRL,x	; stop software reset
	jsr	?20		; wait for 2.5ms
	jsr	wbusy		; wait busy low
	bmi	?04		; device not ready
	asl	a		; N=DRDY status bit
	bpl	?04		; device not ready
	clc
	rts
?04:	lda	#ATA_NOTREADY
	sec			; error
	rts

?20:	lda	#$20		; timer 2.5mS
	trb	VIA0+VIAACR
	lda	#<T2MS5
	sta	VIA0+VIAT2CL
	lda	#>T2MS5
	sta	VIA0+VIAT2CH
?22:	TASKSW	
	lda	#T2IFRB
	bit	VIA0+VIAIFR
	beq	?22
	rts

?40:	lda	#9		; 16uS delay
?42:	dec	a
	bne	?42
	nop
	nop
	nop
	rts

; wait for status bit busy low within BUSYTMO timeout
; in:	X=ata # offset
;	NF=0 if status bit busy low
;	NF=1 if status bit busy high or if timeout
; use:	A
wbusy:
	lda	#$20		; set T2 count PB6 pulse (1ms)
	tsb	VIA0+VIAACR	
	lda	#<BUSYTMO	; timeout status bit busy
	sta	!VIA0+VIAT2CL
	lda	#>BUSYTMO
	sta	!VIA0+VIAT2CH
?10:	lda	!ATAST,x	; read status
	bpl	?20		; OK - busy -> L, return NF = 0
	TASKSW			; switch task while wait
	lda	#T2IFRB
	bit	VIA0+VIAIFR	; test timeout
	beq	?10		; loop
	lda	!ATAST,x	; read status - return NF = busy bit
?20:	rts

; get device identification data
; in:	drive
; out:	CF=1 if error, A=error code
ident:
	ldx	drive
	lda	atadev,x
	and	#11011111B
	sta	atadev,x
	and	#10011111B
	sta	tmpy		; hard disk flag
	txa
	asl	a
	asl	a
	tax
	stx	wtmp		; index for atasec var. access
	ACC16
	stz	atasec,x	; clear atasec var.
	stz	atasec+2,x
	ACC08
	jsr	getid
	bcs	?10		; error
	ldy	#0
	tya
	INDEX16
?02:	clc			; loop checksum computation
	adc	[memp],y
	iny
	cpy	#$200
	bcc	?02
	xba			; B = checksum
	dey
	dey
	lda	[memp],y
	cmp	#$A5		; valid signature checksum?
	INDEX08
	bne	?04		; no -- assume data ok
	xba			; checksum vill be zero
	bne	?15		; wrong checksum
?04:	ldy	#10*2		; offset 'serial number' string
	ldx	#10		; words length
	jsr	?50		; adjust string
	ldy	#23*2		; offset 'firmware revision' string
	ldx	#4		; words length
	jsr	?50		; adjust string
	ldy	#27*2		; offset 'model name' string
	ldx	#20		; words length
	jsr	?50		; adjust string
	ldx	#10*2		; pointer to 'serial number' string
	ldy	#(10*2)+(10*2)-1	; pointer to end of string
	jsr	?60		; terminate string
	ldx	#27*2		; pointer to 'model name' string
	ldy	#(27*2)+(20*2)-1	; pointer to end of string
	jsr	?60		; terminate string
	ldy	#60*2		; word 60/61: total sector's count #
	ldx	wtmp
	ACC16
	lda	[memp],y
	sta	nlba
	sta	atasec,x	; store max. user sectors
	iny
	iny
	lda	[memp],y
	sta	nlba+2
	sta	atasec+2,x
	lda	[memp]		; C.F. identification
	ldx	drive
	ldy	tmpy		; hard disk flag
	cmp	#$848A		; classic word 0 for C.F.
	beq	?06		; is C.F.
	cmp	#$044A		; alternate word 0 for CF
	beq	?06		; is C.F.
	sty	atadev,x	; is H.D.
?06:	ACC08
	lda	atadev,x
	ora	#$20		; flag identify ok
	sta	atadev,x
	clc			; no error
?10:	rts
?15:	lda	#ATA_IDCKS	; checksum error
	sec
	rts

	; swap low byte with high byte in string
	; Y = string offset
	; X = words string length
?50:	ACC16
?52:	lda	[memp],y
	xba
	sta	[memp],y
	iny
	iny
	dex
	bne	?52
	ACC08
	rts

	; terminate serial string and model string with null
	; Y=end of string
	; X=start of string
?60:	stx	tmp	
?62:	lda	[memp],y
	cmp	#' '
	bne	?64
	dey
	cpy	tmp
	bcs	?62
?64:	iny
	lda	#0
	sta	[memp],y
	rts

;--------------------------------------------------------------------------
; init routines
;---------------------------------------------------------------------------

	.LONGA	off
	.LONGI	off

; check & init ata devices
initata:
	.PUBLIC initata

	ACC16
	lda	#!MBR0ATA	; clear ident, MBR & PBR data mem
	sta	bufp
	ldx	#^MBR0ATA
	stx	bufp+2
	ldy	#0
	INDEX16
	tya			; clear A
?02:	sta	[bufp],y
	iny
	iny
	cpy	#$0C00
	bcc	?02
	CPU08
	tax			; device #0
	jsr	?04
	TASKSW
	ldx	#1		; device #1
	jsr	?04
	TASKSW
	rts

; check devices on bus ata
; in:	X=device ata #
?04:	stx	drive
	stz	atadev,x	; clear device flag
	txa
	beq	?06		; offset of ata #0=0
	ldx	#ATAOFS
?06:	stx	ataofs
	jsr	?60		; print start msg
	ldx	ataofs
	jsr	wbusy		; wait for busy flag low
	bmi	?10		; assume no device on ata #0
	jsr	?30		; check device signature
	
	lda	#$00
	sta	!ATADEV,x	; store dev. register

	bcs	?15		; not found or not HD/CF
	
	.IFDEF 	MODE8
	jsr	set8
	.ENDIF
	
	jsr	ident		; identify device
	bcs	?25		; error (A=error code)
	bit	BiosEnt
	bvc	?08		; skip
	jsr	?50		; print identification data
?08:	jsr	devres	
	;jsr	chkmbr		; load&check MBR	
	clc			; no error	
	rts

?10:	jsr	?20		; print 'device not found'
	SCNPRINT
	.DB	'.', 13, 0
	sec
	rts

?15:	jsr	?20		; print 'device not found'
	SCNPRINT
	.DB	' or not H.D./C.F.', 13, 0
	sec
	rts

?20:	SCNPRINT
	.DB	'device not found', $00
	rts

?25:	; fail to identify device
	pha			; error code
	phk
	pea	#!STR2
	lda	#5
	pha
	BPRINTF
	sec
	rts

; check device signature after reset
; in:	drive=ata #
;	ataofs=ata # offset
; out:	CF=0 if found device
;	CF=1 if not device or not HD/CF
?30:	ldy	ataofs
	ldx	#3		; test signature device HD/CF
?34:	lda	>ATASGN,x
	cmp	!ATASCTCNT,y	; test signature
	bne	?36		; no HD/CF
	iny
	dex
	bpl	?34
	lda	!ATASCTCNT,y	; device #
	and	#$10		; mask dev bit
	bne	?36		; will be 0
	iny
	lda	!ATASCTCNT,y	; status reg.
	and	#ATADRDY	; DRDY bit will be set
	beq	?36		; device not found or not HD/CF
	lda	#$C0		; flag device
	ldx	drive		; for now force C.F type
	sta	atadev,x	; found device on ata # 
	clc
	rts
?36:	sec
	rts

?50:	; print identification string
	ldx	drive
	bit	atadev,x
	ACC16CLC
	lda	#!STR3
	bvc	?51
	lda	#!STR4
?51:	sta	wtmp		; H.D or C.F string
	lda	#27*2
	adc	memp
	sta	bufp		; model name string
	lda	nlba
	sta	lba
	lda	nlba+2
	ldx	#11		; divide by 2048 (11 shift)...
?52:	lsr	a		; ... and get rounded size in Mb
	ror	lba
	dex
	bne	?52
	sta	lba+2
	pha			; 2
	lda	lba
	pha			; 2
	ACC08
	lda	memp+2
	pha			; 1
	pei	(bufp)		; 2
	phk			; 1
	pei	(wtmp)		; 2
	phk			; 1
	pea	#!STRF1		; 2
	lda	#14		; 1
	pha
	BPRINTF
	rts

; print check message at boot
; in:	drive=ata # device
?60:	bit	BiosEnt
	bvc	?62		; skip
	ldx	drive
	phx
	phk
	pea	#!STR1
	lda	#5
	pha
	BPRINTF
?62:	rts

; long version
lchkmbr:
	.PUBLIC lchkmbr

	phb
	phd
	pea	#0
	pld
	lda	#0
	pha
	plb
	stx	drive
	jsr	chkmbr
	pld
	plb
	rtl

; check mbr for usb flash disk
chkumbr:
	.PUBLIC chkumbr

	stx	drive
	ACC16			; set pointer for MBR
	lda	#MBR0USB
	sta	memp
	ACC08
	lda	#^MBR0USB
	sta	memp+2
	bra	chkmbr2

; load & check MBR on device
; in:	drive
; out:	CF=1 if error, A=error code
chkmbr:
	jsr	getmbr		; load MBR
	bcc	chkmbr2		
	rts			; error
	
; entry point for usb flash disk
chkmbr2:
	lda	#$10		; set bit mbr loaded
	sta	tmp
	CPU16
	ldy	#$01FE		; signature pointer
	lda	[memp],y
	sta	mbrsgn		; mbr signature
	ldy	#$01C2		; partition type pointer
	lda	[memp],y
	sta	wtmp		; partition type
	ldy	#$01C6		; partition start pointer
	lda	[memp],y
	sta	nlba		; partition start
	iny
	iny
	lda	[memp],y
	sta	nlba+2
	ldy	#$01BE		; partition flag pointer
	lda	[memp],y
	sta	mbrfg		; partition flag
	lda	mbrsgn
	cmp	#$AA55
	CPU08
	bne	?20		; invalid mbr signature
	lda	#$18		; add valid mbr signature flag
	sta	tmp
	lda	mbrfg
	beq	?vv
	bpl	?20		; first partition invalid	
?vv:	lda	#$1C		; add active partition flag
	sta	tmp
	lda	drive
	asl	a
	asl	a
	tax			; index for access to atambr var
	lda	wtmp		; partition type
	tay
	sta	atambr+3,x	; store partition type
	lda	nlba+2		; store partition start
	sta	atambr+2,x
	lda	nlba+1
	sta	atambr+1,x
	lda	nlba
	sta	atambr,x
	lda	nlba+3		; must be 0
	bne	?20
	cpy	#$06		; just FAT16 type
	beq	?10
	cpy	#$04		; FAT 16 for disk less than 32Mb
	beq	?10
	cpy	#$0E
	bne	?20
?10:	CPU16
	ldy	#$01CA		; partition sec's count pointer
	lda	[memp],y
	sta	ataprt,x
	iny
	iny
	lda	[memp],y
	sta	ataprt+2,x
	CPU08
	lda	#$1D		; add valid partition type
	sta	tmp
?20:	ldx	drive		; update device flag
	lda	atadev,x
	ora	tmp
	sta	atadev,x
	clc
?30:	rts

; dma setup for a transfer cycle on channel 1/2
; in:	drive = 0 or 1 (ata #0 or ata #1 device)
;	ataop = ata operation flag: <7>=0->read, 1->write
;	ncnt = words count - 1 #
;	xbuf = dma buffer # ($0000..$07FF)
dmaset:
	; dma buffer is organized in 8 x 128Kb banks or 8 x 64Kw banks
	; PB0..PB2 of VIA3 set MA16..MA18 lines of 512Kw ram (DMA access)
	;
	lda	>DMA0SR		; clear status flag
	sta	>DMA0WMCLR
	
	lda	#$00
	sta	>DMA0WCMD
	
	lda	!VIA3+VIAPRB	; set bank
	and	#11101000B	; mask off 3 lower's bits, clear PB4 (ata 0)   
	ora	xbuf+1		; high byte of buffer # select the bank
	ora	#00100000B	; set /FDC line high (ata dma access)
	ldx	drive
	beq	?via		; select ata 0
	ora	#00010000B	; select ata 1 (PB4 high)
?via:	sta	!VIA3+VIAPRB	; set bank & ata #
	lda	>DMACHNL,x
	sta	wtmp		; dma channel
	ora	#$04
	sta	>DMA0WMSKB	; mask off bit channel 1/2
	lda	>DMAREQB,x
	sta	dmareq
	lda	#WRITETRASF
	bit	ataop
	bpl	?set		; read: dma read ata port and write buffer
	lda	#READTRASF	; write: dma read buffer and write to ata
?set:	ora	wtmp		; channel -  A=transfer mode
	sta	>DMA0MODE	; set DMA transfer mode
	;lda	xbuf		; high address
	txy
	sta	>DMA0WCLRFF	; clear F/F
	bne	?chn2		; set channel 2
	
	; set channel 1
	lda	#0
	sta	>DMAC0+DMAADDR1	; store low buffer address
	lda	xbuf		; high address
	sta	>DMAC0+DMAADDR1	; store high buffer address
	nop
	sta	>DMA0WCLRFF	; clear F/F
	lda	ncnt		; low count
	sta	>DMAC0+DMACNT1	; store low count
	lda	ncnt+1		; high count
	sta	>DMAC0+DMACNT1	; store high count
	bra	?cnt
	
	; set channel 2
?chn2:	lda	#0
	sta	>DMAC0+DMAADDR2	; store low buffer address
	lda	xbuf		; high address
	sta	>DMAC0+DMAADDR2	; store high buffer address
	nop
	sta	>DMA0WCLRFF	; clear F/F
	lda	ncnt		; low count
	sta	>DMAC0+DMACNT2	; store low count
	lda	ncnt+1		; high count
	sta	>DMAC0+DMACNT2	; store high count
?cnt:	sta	>DMA0WCLRFF	; clear F/F	
	lda	wtmp
	sta	>DMA0WMSKB	; un-mask bit channel 1/2
	lda	#00000100B	; VIA0 PA3 set dma direction
	bit	ataop
	bmi	?dir		; clear direction (write to ata)
	tsb	VIA0+VIAPRANH	; set direction (write to mem)
	rts
?dir:	trb	VIA0+VIAPRANH	; clear direction
	rts
	
;--------------------------------------------------------------------------
; tables
;---------------------------------------------------------------------------

; operation flag (read, verify, write)
OPTAB:
	.DB	$00, $40, $80

; dma operations: read, write
OPTAB2:
	.DB	$00, $80

; dma channel related to drive #
DMACHNL:
	.DB	$01, $02

; dma channel request bit related to drive #
DMAREQB:
	.DB	$20, $40

; DMAREQ bit (PA7, PA6)
DMAREQ:
	.DB	$80, $40

; for test VIA1 IFR (interrupt from ata device)
IFRTST:
	.DB	CA1IFRB, CA2IFRB

; HD/CF signature (in reverse order)	
ATASGN:
	.DB	$00, $00, $01, $01

STR1:
	.DB	'ata   #%bu...', 0

STR2:
	.DB	'fail to identify device (error %bu).', 13, 0

STR3:
	.DB	'Hard Disk    ', 0

STR4:
	.DB	'Compact Flash', 0
	
STRF1:
	.DB	'%s %-42.41s %,7u Mb', 13, 0

; string's address (drive # indexed)
DSKMODEL:
	.DB	<_atamod0, <_atamod1, <_usbdsc0, <_usbdsc1

DSKMODELH:
	.DB	>_atamod0, >_atamod1, >_usbdsc0, >_usbdsc1
	
DSKSER:
	.DB	<_ataser0, <_ataser1

DSKSERH:
	.DB	>_ataser0, >_ataser1
	
DSKREV:
	.DB	<_atarev0, <_atarev1

DSKREVH:
	.DB	>_atarev0, >_atarev1
